@@ -2,7 +2,7 @@
  * nvidia-persistenced: A daemon for maintaining persistent driver state,
  * specifically for use by the NVIDIA Linux driver.
  *
- * Copyright (C) 2013-2014 NVIDIA Corporation
+ * Copyright (C) 2013-2018 NVIDIA Corporation
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -33,18 +33,16 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <syslog.h>
 #include <sys/socket.h>
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <unistd.h>
 
-#include "common-utils.h"
 #include "nvidia-persistenced.h"
-#include "nvidia-cfg.h"
 #include "nvpd_defs.h"
 #include "nvpd_rpc.h"
-
+#include "nvidia-numa.h"
+#include "nvidia-syslog-utils.h"
 /*
  * Local Definitions
  */
@@ -56,6 +54,8 @@ typedef struct
     NvCfgDeviceHandle nv_cfg_handle;
     NvCfgPciDevice pci_info;
     NvPersistenceMode mode;
+    NvNumaStatus numa_status;
+    NvNumaDevice numa_info;
 } NvPdDevice;
 
 /*
@@ -67,9 +67,7 @@ static int pid_fd = -1;
 static int socket_fd = -1;
 static NvPdDevice *devices = NULL;
 static int num_devices = 0;
-static int log_mask = 0;
 static int remove_dir = 0;
-static int verbose = 0;
 
 static struct {
     NvCfgBool (*get_pci_devices)(int *, NvCfgPciDevice **);
@@ -80,14 +78,15 @@ static struct {
 /*
  * Local Functions
  */
-static void daemonize(uid_t uid, gid_t gid);
+static int daemonize(uid_t uid, gid_t gid);
 static int load_nvidia_cfg_sym(void **sym_ptr, const char *sym_name);
+static NvPdDevice *get_device(int domain, int bus, int slot);
 static NvPdStatus setup_nvidia_cfg_api(const char *nvidia_cfg_path);
 static NvPdStatus setup_devices(NvPersistenceMode default_mode);
 static NvPdStatus setup_rpc(void);
 static NvPdStatus set_device_mode(NvPdDevice *device, NvPersistenceMode mode);
-static void syslog_device(NvPdDevice *device, int priority,
-                          const char *format, ...);
+static NvPdStatus set_device_numa_status(NvPdDevice *device,
+                                         NvNumaStatus numa_status);
 
 /*
  * nvPdSetDevicePersistenceMode() - This function implements the daemon
@@ -100,25 +99,77 @@ static void syslog_device(NvPdDevice *device, int priority,
 NvPdStatus nvPdSetDevicePersistenceMode(int domain, int bus, int slot,
                                         int function, NvPersistenceMode mode)
 {
-    int i = 0;
-    NvPdStatus ret = NVPD_ERR_DEVICE_NOT_FOUND;
+    NvPdStatus ret;
+    NvPersistenceMode old_mode;
+    NvPdDevice *device = get_device(domain, bus, slot);
 
-    if (devices == NULL) {
-        return ret;
+    if (device == NULL) {
+        return NVPD_ERR_DEVICE_NOT_FOUND;
     }
 
-    for (i = 0; i < num_devices; i++) {
-        if ((devices[i].pci_info.domain == domain) &&
-            (devices[i].pci_info.bus == bus) &&
-            (devices[i].pci_info.slot == slot)) {
-            ret = set_device_mode(&devices[i], mode);
-            break;
+    old_mode = device->mode;
+
+    /*
+     * Set the device mode always before changing the NUMA state.
+     * For onlining, this is needed since libnvidia-cfg needs to create the
+     * device nodes before nvidia-numa can interact with them.
+     * For offlining, this is needed since the libnvidia-cfg device handle
+     * will need to be released for nvidia-numa to proceed with the offlining.
+     */
+    ret = set_device_mode(device, mode);
+    if (ret == NVPD_SUCCESS) {
+        NvNumaStatus status = (mode == NV_PERSISTENCE_MODE_ENABLED) ?
+                                NV_NUMA_STATUS_ONLINE : NV_NUMA_STATUS_OFFLINE;
+        ret = set_device_numa_status(device, status);
+        if ((ret != NVPD_SUCCESS) && (old_mode != mode)) {
+            (void) set_device_mode(device, old_mode);
         }
     }
 
     return ret;
 }
 
+/*
+ * nvPdSetDevicePersistenceModeOnly() - This function implements the daemon
+ * command to set the persistence mode of the device at the specified PCI
+ * location, without affecting the NUMA status of the device.
+ *
+ * The function parameter is ignored for the time being, and provided for
+ * completeness of the API.
+ */
+NvPdStatus nvPdSetDevicePersistenceModeOnly(int domain, int bus, int slot,
+                                            int function,
+                                            NvPersistenceMode mode)
+{
+    NvPdDevice *device = get_device(domain, bus, slot);
+
+    if (device == NULL) {
+        return NVPD_ERR_DEVICE_NOT_FOUND;
+    }
+
+    return set_device_mode(device, mode);
+}
+
+/*
+ * nvPdSetDeviceNumaStatus() - This function implements the daemon command to
+ * set the NUMA status of the device at the specified PCI location, without
+ * affecting the persistence mode of the device.
+ *
+ * The function parameter is ignored for the time being, and provided for
+ * completeness of the API.
+ */
+NvPdStatus nvPdSetDeviceNumaStatus(int domain, int bus, int slot, int function,
+                                   NvNumaStatus status)
+{
+    NvPdDevice *device = get_device(domain, bus, slot);
+
+    if (device == NULL) {
+        return NVPD_ERR_DEVICE_NOT_FOUND;
+    }
+
+    return set_device_numa_status(device, status);
+}
+
 /*
  * nvPdGetDevicePersistenceMode() - This function implements the daemon
  * command to get the persistence mode of the device at the specified PCI
@@ -130,63 +181,87 @@ NvPdStatus nvPdSetDevicePersistenceMode(int domain, int bus, int slot,
 NvPdStatus nvPdGetDevicePersistenceMode(int domain, int bus, int slot,
                                         int function, NvPersistenceMode *mode)
 {
-    int i = 0;
-    NvPdStatus ret = NVPD_ERR_DEVICE_NOT_FOUND;
+    NvPdDevice *device = get_device(domain, bus, slot);
 
-    if (devices == NULL) {
-        return ret;
+    if (device == NULL) {
+        return NVPD_ERR_DEVICE_NOT_FOUND;
     }
 
-    for (i = 0; i < num_devices; i++) {
-        if ((devices[i].pci_info.domain == domain) &&
-            (devices[i].pci_info.bus == bus) &&
-            (devices[i].pci_info.slot == slot)) {
-            *mode = devices[i].mode;
-            ret = NVPD_SUCCESS;
-            break;
-        }
-    }
-
-    return ret;
+    *mode = device->mode;
+    return NVPD_SUCCESS;
 }
 
 /*
- * syslog_device() - This function prints the device info along with the
- * specified message to syslog. If there is a failure to allocate memory along
- * the way, it will simply fail with an error to syslog().
+ * get_device() - looks up and returns a pointer to the NvPdDevice structure
+ * for the device at the specified PCI location.
  */
-static void syslog_device(NvPdDevice *device, int priority,
-                          const char *format, ...)
+static NvPdDevice *get_device(int domain, int bus, int slot)
 {
-    char *device_str = NULL;
-    char *str = NULL;
+    int i;
 
-    /* First check if this message will even show up. */
-    if ((log_mask & priority) == 0) {
-        return;
+    if (devices != NULL) {
+        for (i = 0; i < num_devices; i++) {
+            if ((devices[i].pci_info.domain == domain) &&
+                (devices[i].pci_info.bus == bus) &&
+                (devices[i].pci_info.slot == slot)) {
+                return &devices[i];
+            }
+        }
     }
 
-    /* First fill in the format string as usual. */
-    NV_VSNPRINTF(str, format);
-    if (str == NULL) {
-        syslog(LOG_ERR, "Failed to create formatted message.");
-        return;
+    return NULL;
+}
+
+/*
+ * init_complete() - called by the child (daemon) process to signal to the
+ * parent process, via the init pipe created during daemonize(), that
+ * initialization has completed successfully.
+ */
+static NvPdStatus init_complete(int pipe_write_fd)
+{
+    unsigned char success = 1;
+    int bytes;
+    
+    bytes = write(pipe_write_fd, &success, sizeof(success));
+    
+    close(pipe_write_fd);
+
+    if (bytes < 0) {
+        fprintf(stderr, "Failed to write init pipe: %s\n", strerror(errno));
+        return NVPD_ERR_IO;
     }
 
-    /* Then prefix it with the device info. */
-    device_str = nvasprintf("device %04d:%02x:%02x.%x - %s",
-                            device->pci_info.domain, device->pci_info.bus,
-                            device->pci_info.slot, device->pci_info.function,
-                            str);
-    nvfree(str);
-    if (device_str == NULL) {
-        syslog(LOG_ERR, "Failed to create device message.");
-        return;
+    return NVPD_SUCCESS;
+}
+
+/*
+ * wait_for_init_complete() - called by the parent process to block on the
+ * init pipe and wait for the child process to signal its successful
+ * initialization.
+ *
+ * The init pipe will be closed upon return from this function.
+ */
+static NvPdStatus wait_for_init_complete(int pipe_read_fd)
+{
+    unsigned char success = 0;
+    int bytes;
+
+    bytes = read(pipe_read_fd, &success, sizeof(success));
+
+    close(pipe_read_fd);    
+
+    if (bytes < 0) {
+        fprintf(stderr, "Failed to read init pipe: %s\n", strerror(errno));
+        return NVPD_ERR_IO;
     }
 
-    syslog(priority, "%s", device_str);
+    if (bytes != sizeof(success) || !success) {
+        fprintf(stderr, "nvidia-persistenced failed to initialize. "
+                        "Check syslog for more details.\n");
+        return NVPD_ERR_UNKNOWN;
+    }
 
-    nvfree(device_str);
+    return NVPD_SUCCESS;
 }
 
 /*
@@ -201,30 +276,23 @@ static NvPdStatus set_device_mode(NvPdDevice *device, NvPersistenceMode mode)
 
     /* If the device is already in the mode specified, just abort */
     if (mode == device->mode) {
-        if (verbose) {
-            syslog_device(device, LOG_NOTICE,
-                          "already in requested persistence mode.");
-        }
+        SYSLOG_DEVICE_VERBOSE(&device->pci_info, LOG_NOTICE,
+                              "already in requested persistence mode.");
         return status;
     }
 
     switch (mode) {
 
     case NV_PERSISTENCE_MODE_DISABLED:
-
         /* If the new mode is disabled, we must detach the device. */
         success = nv_cfg_api.detach_device(device->nv_cfg_handle);
         if (!success) {
-            syslog_device(device, LOG_ERR, "failed to detach.");
+            syslog_device(&device->pci_info, LOG_ERR, "failed to detach.");
             status = NVPD_ERR_DRIVER;
         } else {
-            if (verbose) {
-                syslog_device(device, LOG_NOTICE,
-                              "persistence mode disabled.");
-            }
             device->nv_cfg_handle = NULL;
         }
-    
+
         break;
 
     case NV_PERSISTENCE_MODE_ENABLED:
@@ -236,17 +304,16 @@ static NvPdStatus set_device_mode(NvPdDevice *device, NvPersistenceMode mode)
                                                device->pci_info.function,
                                                &device->nv_cfg_handle);
         if (!success) {
-            syslog_device(device, LOG_ERR, "failed to attach.");
+            syslog_device(&device->pci_info, LOG_ERR, "failed to attach.");
             status = NVPD_ERR_DRIVER;
-        } else if (verbose) {
-            syslog_device(device, LOG_NOTICE, "persistence mode enabled.");
         }
 
         break;
 
     default:
 
-        syslog_device(device, LOG_ERR, "requested invalid mode %d", mode);
+        syslog_device(&device->pci_info, LOG_ERR,
+                      "requested invalid mode %d", mode);
         status = NVPD_ERR_INVALID_ARGUMENT;
         break;
 
@@ -254,6 +321,68 @@ static NvPdStatus set_device_mode(NvPdDevice *device, NvPersistenceMode mode)
 
     if (status == NVPD_SUCCESS) {
         device->mode = mode;
+        SYSLOG_DEVICE_VERBOSE(&device->pci_info, LOG_NOTICE,
+                              "persistence mode %s.",
+                              (mode == NV_PERSISTENCE_MODE_ENABLED) ?
+                                "enabled" : "disabled");
+    }
+
+    return status;
+}
+
+/*
+ * set_device_numa_status() - This function invokes the nvidia-numa functions
+ * for onlining or offlining device NUMA memory for a given device.
+ */
+static NvPdStatus set_device_numa_status(NvPdDevice *device,
+                                         NvNumaStatus numa_status)
+{
+    NvPdStatus status = NVPD_SUCCESS;
+
+    /* If the device is already in the state specified, just abort */
+    if (numa_status == device->numa_status) {
+        SYSLOG_DEVICE_VERBOSE(&device->pci_info, LOG_NOTICE,
+                              "NUMA memory already in requested state.");
+        return status;
+    }
+
+    switch (numa_status) {
+
+    case NV_NUMA_STATUS_OFFLINE:
+
+        status = nvNumaOfflineMemory(&device->numa_info);
+        if (status != NVPD_SUCCESS) {
+            syslog_device(&device->pci_info, LOG_ERR,
+                          "failed to offline memory.\n");
+        }
+
+        break;
+
+    case NV_NUMA_STATUS_ONLINE:
+
+        status = nvNumaOnlineMemory(&device->numa_info);
+        if (status != NVPD_SUCCESS) {
+            syslog_device(&device->pci_info, LOG_ERR,
+                          "failed to online memory.\n");
+        }
+
+        break;
+
+    default:
+
+        syslog_device(&device->pci_info, LOG_ERR,
+                      "requested invalid NUMA status %d", numa_status);
+        status = NVPD_ERR_INVALID_ARGUMENT;
+        break;
+
+    }
+
+    if (status == NVPD_SUCCESS) {
+        device->numa_status = numa_status;
+        SYSLOG_DEVICE_VERBOSE(&device->pci_info, LOG_NOTICE,
+                              "NUMA memory %s.",
+                              (numa_status == NV_NUMA_STATUS_ONLINE) ?
+                                "onlined" : "offlined");
     }
 
     return status;
@@ -269,16 +398,22 @@ static void shutdown_daemon(int status)
 {
     int i;
 
+    /* Nothing to clean up */
+    if (pid <= 0) {
+        goto shutdown;
+    }
+
     /* Clean up and remove the RPC socket */
     if (socket_fd != -1) {
-        /* Unregister any mappings to the RPC dispatch routine */
+        /* Unregister any mappings to the RPC dispatch routines */
         svc_unregister(NVPD_PROG, VersionOne);
+        svc_unregister(NVPD_PROG, VersionTwo);
 
         if (close(socket_fd) < 0) {
             syslog(LOG_ERR, "Failed to close socket: %s",
                    strerror(errno));
-        } else if (verbose) {
-            syslog(LOG_INFO, "Socket closed.");
+        } else {
+            SYSLOG_VERBOSE(LOG_INFO, "Socket closed.");
         }
 
         if (unlink(NVPD_SOCKET_PATH) < 0) {
@@ -291,8 +426,11 @@ static void shutdown_daemon(int status)
     if (devices != NULL) {
         for (i = 0; i < num_devices; i++) {
             if (devices[i].nv_cfg_handle != NULL) {
-                (void) set_device_mode(&devices[i],
-                                       NV_PERSISTENCE_MODE_DISABLED);
+                NvPersistenceMode mode = NV_PERSISTENCE_MODE_DISABLED;
+                (void) nvPdSetDevicePersistenceMode(devices[i].pci_info.domain,
+                                                    devices[i].pci_info.bus,
+                                                    devices[i].pci_info.slot,
+                                                    0, mode);
             }
         }
 
@@ -311,16 +449,16 @@ static void shutdown_daemon(int status)
         if (lockf(pid_fd, F_ULOCK, 0) < 0) {
             syslog(LOG_ERR, "Failed to unlock PID file: %s",
                    strerror(errno));
-        } else if (verbose) {
-            syslog(LOG_INFO, "PID file unlocked.");
+        } else {
+            SYSLOG_VERBOSE(LOG_INFO, "PID file unlocked.");
         }
 
         /* Close the PID file */
         if (close(pid_fd) < 0) {
             syslog(LOG_ERR, "Failed to close PID file: %s",
                    strerror(errno));
-        } else if (verbose) {
-            syslog(LOG_INFO, "PID file closed.");
+        } else {
+            SYSLOG_VERBOSE(LOG_INFO, "PID file closed.");
         }
 
         /* Remove the PID file */
@@ -330,7 +468,7 @@ static void shutdown_daemon(int status)
         }
     }
 
-    /* 
+    /*
      * Remove the runtime data directory if the daemon created it. If the
      * daemon has dropped permissions and is no longer able to remove the
      * directory, issue a notice instead of a warning, as this is expected.
@@ -338,12 +476,10 @@ static void shutdown_daemon(int status)
     if (remove_dir && (rmdir(NVPD_VAR_RUNTIME_DATA_PATH) < 0) &&
         (errno != ENOENT)) {
         if (errno == EACCES) {
-            if (verbose) {
-                syslog(LOG_NOTICE,
-                       "The daemon no longer has permission to remove its "
-                       "runtime data directory %s",
-                       NVPD_VAR_RUNTIME_DATA_PATH);
-            }
+            SYSLOG_VERBOSE(LOG_NOTICE,
+                           "The daemon no longer has permission to remove its "
+                           "runtime data directory %s",
+                           NVPD_VAR_RUNTIME_DATA_PATH);
         } else {
             syslog(LOG_WARNING, "Failed to remove runtime data directory: %s",
                    strerror(errno));
@@ -353,6 +489,7 @@ static void shutdown_daemon(int status)
     syslog(LOG_NOTICE, "Shutdown (%d)", pid);
     closelog();
 
+shutdown:
     exit(status);
 }
 
@@ -392,7 +529,7 @@ static NvPdStatus setup_nvidia_cfg_api(const char *nvidia_cfg_path)
     if (nvidia_cfg_path != NULL) {
         nvfree(lib_path);
     }
-    
+
     if (libnvidia_cfg == NULL) {
         syslog(LOG_ERR, "Failed to open %s: %s", NVIDIA_CFG_LIB,
                dlerror());
@@ -456,12 +593,18 @@ static NvPdStatus setup_devices(NvPersistenceMode default_mode)
         devices[i].pci_info.function = 0;
         devices[i].mode = NV_PERSISTENCE_MODE_DISABLED;
 
-        if (verbose) {
-            syslog_device(&devices[i], LOG_DEBUG, "registered");
-        }
+        /* Initialize nvidia-numa state */
+        devices[i].numa_status = NV_NUMA_STATUS_OFFLINE;
+        devices[i].numa_info.fd = -1;
+        devices[i].numa_info.pci_info = &devices[i].pci_info;
+
+        SYSLOG_DEVICE_VERBOSE(&(devices[i].pci_info), LOG_DEBUG, "registered");
 
         if (default_mode != NV_PERSISTENCE_MODE_DISABLED) {
-            (void) set_device_mode(&devices[i], default_mode);
+            (void) nvPdSetDevicePersistenceMode(devices[i].pci_info.domain,
+                                                devices[i].pci_info.bus,
+                                                devices[i].pci_info.slot, 0,
+                                                default_mode);
         }
     }
 
@@ -503,14 +646,17 @@ static NvPdStatus setup_rpc()
     }
 
     if (!svc_register(transp, NVPD_PROG, VersionOne, nvpd_prog_1, 0)) {
-        syslog(LOG_ERR, "Failed to register RPC service");
+        syslog(LOG_ERR, "Failed to register RPC V1 service");
         return NVPD_ERR_RPC;
     }
 
-    if (verbose) {
-        syslog(LOG_INFO, "Local RPC service initialized");
+    if (!svc_register(transp, NVPD_PROG, VersionTwo, nvpd_prog_2, 0)) {
+        syslog(LOG_ERR, "Failed to register RPC V2 service");
+        return NVPD_ERR_RPC;
     }
 
+    SYSLOG_VERBOSE(LOG_INFO, "Local RPC services initialized");
+
     return NVPD_SUCCESS;
 }
 
@@ -520,9 +666,7 @@ static NvPdStatus setup_rpc()
  */
 static void signal_handler(int signal)
 {
-    if (verbose) {
-        syslog(LOG_DEBUG, "Received signal %d", signal);
-    }
+    SYSLOG_VERBOSE(LOG_DEBUG, "Received signal %d", signal);
 
     switch (signal) {
 
@@ -542,13 +686,14 @@ static void signal_handler(int signal)
  * daemonize() - This function converts the current process into a daemon
  * process.
  */
-static void daemonize(uid_t uid, gid_t gid)
+static int daemonize(uid_t uid, gid_t gid)
 {
     char pid_str[10];
     struct sigaction signal_action;
     sigset_t signal_set;
+    int init_pipe_fds[2];
+    int pipe_read_fd, pipe_write_fd;
     int fd;
-    int status = EXIT_SUCCESS;
 
     /*
      * Set up the signal handler - block TTY-related signals, catch
@@ -568,35 +713,61 @@ static void daemonize(uid_t uid, gid_t gid)
     sigaction(SIGINT,  &signal_action, NULL);
     sigaction(SIGTERM, &signal_action, NULL);
 
-    pid = fork();
-    if (pid < 0) {
-        syslog(LOG_ERR, "Failed to fork() daemon: %s",
-               strerror(errno));
+    /*
+     * Set up the init pipe for coordinating daemon init with main process
+     * return.
+     */
+    if (pipe(init_pipe_fds) < 0) {
+        fprintf(stderr, "Failed to create init pipe: %s\n", strerror(errno));
         exit(EXIT_FAILURE);
     }
-    else if (pid > 0) {
-        /* Kill off the parent process */
-        exit(EXIT_SUCCESS);
-    }
 
-    /* Save off the new pid for logging */
-    pid = getpid();
+    pipe_read_fd = init_pipe_fds[0];
+    pipe_write_fd = init_pipe_fds[1];
+
+    pid = fork();
+    if (pid < 0) {
+        fprintf(stderr, "Failed to fork() daemon: %s", strerror(errno));
+        goto shutdown;
+    }
+    else if (pid > 0) {
+        NvPdStatus init_status;
+
+        /*
+         * Close the write end of the pipe so we don't block if the child dies
+         * or otherwise closes its write fd before it sends the init message.
+         */
+        close(pipe_write_fd);
+
+        /*
+         * Watch the init pipe for the child process to init, then exit the
+         * parent process.
+         */
+        init_status = wait_for_init_complete(pipe_read_fd);
+        exit((init_status == NVPD_SUCCESS) ? EXIT_SUCCESS : EXIT_FAILURE);
+    }
 
     /* Reset default file permissions */
     umask(0);
 
     /* Create a new session for the daemon */
     if (setsid() < 0) {
-        syslog(LOG_ERR, "Failed to create new daemon session: %s",
-               strerror(errno));
-        exit(EXIT_FAILURE);
+        fprintf(stderr, "Failed to create new daemon session: %s",
+                strerror(errno));
+        goto shutdown;
     }
 
+    /* Save off the new pid for logging */
+    pid = getpid();
+
     /* Close the standard file descriptors */
     close(STDIN_FILENO);
     close(STDOUT_FILENO);
     close(STDERR_FILENO);
 
+    /* Close the read end of the init pipe */
+    close(pipe_read_fd);
+
     if (verbose) {
         log_mask = LOG_UPTO(LOG_DEBUG);
     } else {
@@ -607,9 +778,7 @@ static void daemonize(uid_t uid, gid_t gid)
 
     /* Setup syslog connection */
     openlog(NVPD_DAEMON_NAME, 0, LOG_DAEMON);
-    if (verbose) {
-        syslog(LOG_INFO, "Verbose syslog connection opened");
-    }
+    SYSLOG_VERBOSE(LOG_INFO, "Verbose syslog connection opened");
 
     /* Go somewhere that we won't be unmounted */
     if (chdir("/") < 0) {
@@ -617,7 +786,7 @@ static void daemonize(uid_t uid, gid_t gid)
                strerror(errno));
     }
 
-    /* 
+    /*
      * Try to create the supplied data path - if it fails, we'll catch the
      * error with the access() call below.
      */
@@ -627,12 +796,10 @@ static void daemonize(uid_t uid, gid_t gid)
                    NVPD_VAR_RUNTIME_DATA_PATH, strerror(errno));
         }
 
-        if (verbose) {
-            syslog(LOG_INFO, "Directory %s will not be removed on exit",
-                   NVPD_VAR_RUNTIME_DATA_PATH);
-        }
+        SYSLOG_VERBOSE(LOG_INFO, "Directory %s will not be removed on exit",
+                       NVPD_VAR_RUNTIME_DATA_PATH);
     } else {
-        /* 
+        /*
          * Only attempt to remove the directory on shutdown if the daemon
          * created it.
          */
@@ -647,26 +814,21 @@ static void daemonize(uid_t uid, gid_t gid)
         if (chown(NVPD_VAR_RUNTIME_DATA_PATH, uid, gid) < 0) {
             syslog(LOG_ERR, "Failed to change ownership of %s: %s",
                    NVPD_VAR_RUNTIME_DATA_PATH, strerror(errno));
-            status = EXIT_FAILURE;
-            goto done;
+            goto shutdown;
         }
 
         if (setgid(gid) < 0) {
             syslog(LOG_ERR, "Failed to set group ID: %s", strerror(errno));
-            status = EXIT_FAILURE;
-            goto done;
+            goto shutdown;
         }
 
         if (setuid(uid) < 0) {
             syslog(LOG_ERR, "Failed to set user ID: %s", strerror(errno));
-            status = EXIT_FAILURE;
-            goto done;
+            goto shutdown;
         }
 
-        if (verbose) {
-            syslog(LOG_INFO, "Now running with user ID %d and group ID %d",
-                   uid, gid);
-        }
+        SYSLOG_VERBOSE(LOG_INFO, "Now running with user ID %d and group ID %d",
+                       uid, gid);
     }
 
     /*
@@ -675,8 +837,7 @@ static void daemonize(uid_t uid, gid_t gid)
     if (access(NVPD_VAR_RUNTIME_DATA_PATH, R_OK | W_OK) < 0) {
         syslog(LOG_ERR, "Unable to access %s: %s",
                NVPD_VAR_RUNTIME_DATA_PATH, strerror(errno));
-        status = EXIT_FAILURE;
-        goto done;
+        goto shutdown;
     }
 
     /*
@@ -686,16 +847,14 @@ static void daemonize(uid_t uid, gid_t gid)
     fd = open(NVPD_PID_FILE, O_RDWR | O_CREAT, 0644);
     if (fd < 0) {
         syslog(LOG_ERR, "Failed to open PID file: %s", strerror(errno));
-        status = EXIT_FAILURE;
-        goto done;
+        goto shutdown;
     }
 
     /* Lock the PID file */
     if (lockf(fd, F_TLOCK, 0) < 0) {
         syslog(LOG_ERR, "Failed to lock PID file: %s", strerror(errno));
-        status = EXIT_FAILURE;
         close(fd);
-        goto done;
+        goto shutdown;
     }
 
     /*
@@ -703,54 +862,68 @@ static void daemonize(uid_t uid, gid_t gid)
      */
     pid_fd = fd;
 
-done:
-    if (status == EXIT_SUCCESS) {
-        /* Update the PID file with the current process ID */
-        sprintf(pid_str, "%d\n", pid);
-        if (write(pid_fd, pid_str, strlen(pid_str)) != strlen(pid_str)) {
-            syslog(LOG_ERR, "Failed to update PID file: %s", strerror(errno));
-            shutdown_daemon(EXIT_FAILURE);
-        }
-        syslog(LOG_NOTICE, "Started (%d)", pid);
-    }
-    else {
-        /*
-         * If we get this far but have an error condition, we need to cleanup
-         * any runtime files left around.
-         */
-        shutdown_daemon(status);
+    /* Update the PID file with the current process ID */
+    sprintf(pid_str, "%d\n", pid);
+    if (write(pid_fd, pid_str, strlen(pid_str)) != strlen(pid_str)) {
+        syslog(LOG_ERR, "Failed to update PID file: %s", strerror(errno));
+        goto shutdown;
     }
+
+    syslog(LOG_NOTICE, "Started (%d)", pid);
+    return pipe_write_fd;
+
+shutdown:
+    close(pipe_write_fd);
+
+    /*
+     * If we get this far but have an error condition, we need to cleanup
+     * any runtime files left around.
+     */
+    shutdown_daemon(EXIT_FAILURE);
+
+    /* Unreachable */
+    return -1;
 }
 
 int main(int argc, char* argv[])
 {
     NvPdStatus status;
     NvPdOptions options;
+    int pipe_write_fd;
 
     parse_options(argc, argv, &options);
     verbose = options.verbose;
 
-    daemonize(options.uid, options.gid);
+    pipe_write_fd = daemonize(options.uid, options.gid);
 
+    /* Only the daemon process reaches this point */
     status = setup_nvidia_cfg_api(options.nvidia_cfg_path);
     if (status != NVPD_SUCCESS) {
-        shutdown_daemon(EXIT_FAILURE);
+        goto shutdown;
     }
 
     status = setup_devices(options.persistence_mode);
     if (status != NVPD_SUCCESS) {
-        shutdown_daemon(EXIT_FAILURE);
+        goto shutdown;
     }
 
     status = setup_rpc();
     if (status != NVPD_SUCCESS) {
-        shutdown_daemon(EXIT_FAILURE);
+        goto shutdown;
+    }
+
+    status = init_complete(pipe_write_fd);
+    if (status != NVPD_SUCCESS) {
+        goto shutdown;
     }
 
     svc_run();
 
     /* We should never return from svc_run() in a non-error scenario */
     syslog(LOG_ERR, "Failed to start local RPC service");
+
+shutdown:
+    close(pipe_write_fd);
     shutdown_daemon(EXIT_FAILURE);
 
     return 0;
